pub(crate) mod bit_operations;
pub(crate) mod builder_struct;
pub(crate) mod common;
pub(crate) mod debug_impl;
pub(crate) mod default_impl;
pub(crate) mod field_const_getter_setter;
pub(crate) mod from_into_bits_conversions;
pub(crate) mod from_types_impl;
pub(crate) mod new_impl;
pub(crate) mod set_clear_bits_impl;
pub(crate) mod tuple_struct;
